/////////////////////////////////////////////////////////////////////////////
//
// Windows MFC Glk Libraries
//
// GlkSound
// Glk sound channels
//
/////////////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "GlkDll.h"
#include "GlkSoundAIFF.h"
#include "GlkSoundMOD.h"

#include <math.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// Class for Glk sound channels
/////////////////////////////////////////////////////////////////////////////

IMPLEMENT_DYNAMIC(CWinGlkSndChannel,CObject);

CWinGlkSndChannel::CWinGlkSndChannel(glui32 Rock) : CObject()
{
	m_Rock = Rock;
	m_pSound = NULL;
	m_iSound = 0;
	m_iVolume = 0x10000;
	m_iNotify = 0;

	// Add to the global map of all sound channels
	ChannelMap[this] = 0;

	if (RegisterObjFn)
		SetDispRock((*RegisterObjFn)(this,gidisp_Class_Schannel));
	else
		m_DispRock.num = 0;

	// Make sure that the sound engine is set up
	((CGlkApp*)AfxGetApp())->InitSoundEngine();
}

CWinGlkSndChannel::~CWinGlkSndChannel()
{
	Stop();

	if (UnregisterObjFn)
		(*UnregisterObjFn)(this,gidisp_Class_Schannel,GetDispRock());

	// Remove from the global map
	ChannelMap.RemoveKey(this);
}

glui32 CWinGlkSndChannel::GetRock(void)
{
	return m_Rock;
}

bool CWinGlkSndChannel::Play(CWinGlkSound* pSound, glui32 iSound, int iRepeat, int iNotify)
{
	Stop();

	// Store the new sound data
	m_pSound = pSound;
	m_iSound = iSound;
	m_iNotify = iNotify;

	if (m_pSound->Play(iRepeat,m_iVolume) == false)
	{
		Stop();
		return false;
	}
	return true;
}

void CWinGlkSndChannel::Stop(void)
{
	if (m_pSound != NULL)
		delete m_pSound;
	m_pSound = NULL;

	m_iSound = 0;
	m_iNotify = 0;
}

void CWinGlkSndChannel::SetVolume(int iVolume)
{
	m_iVolume = iVolume;
	if (m_pSound != NULL)
		m_pSound->SetVolume(iVolume);
}

void CWinGlkSndChannel::TimerPulse(void)
{
	if (m_pSound != NULL)
	{
		// Is the sound still playing?
		if (m_pSound->IsPlaying() == false)
		{
			// Post a sound notification event
			if (m_iNotify != 0)
				((CGlkApp*)AfxGetApp())->AddEvent(evtype_SoundNotify,0,m_iSound,m_iNotify);
			Stop();
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
// Static data and member functions

CMap<CWinGlkSndChannel*,CWinGlkSndChannel*,int,int> CWinGlkSndChannel::ChannelMap;

void CWinGlkSndChannel::CloseAllChannels(void)
{
	CWinGlkSndChannel* pChannel;
	CArray<CWinGlkSndChannel*,CWinGlkSndChannel*> ChannelArray;
	int i;

	POSITION MapPos = ChannelMap.GetStartPosition();
	while (MapPos)
	{
		ChannelMap.GetNextAssoc(MapPos,pChannel,i);
		ChannelArray.Add(pChannel);
	}

	for (i = 0; i < ChannelArray.GetSize(); i++)
		delete ChannelArray[i];
}

bool CWinGlkSndChannel::IsValidChannel(CWinGlkSndChannel* pChannel)
{
	int iDummy;
	if (ChannelMap.Lookup(pChannel,iDummy))
		return true;
	return false;
}

CWinGlkSndChannel* CWinGlkSndChannel::IterateChannels(CWinGlkSndChannel* pChannel, glui32* pRockPtr)
{
	CWinGlkSndChannel* pMapChannel = NULL;
	CWinGlkSndChannel* pPrevChannel = NULL;
	int iDummy;

	POSITION MapPos = ChannelMap.GetStartPosition();
	while (MapPos)
	{
		ChannelMap.GetNextAssoc(MapPos,pMapChannel,iDummy);
		if (pChannel == pPrevChannel)
		{
			if (pRockPtr)
				*pRockPtr = pMapChannel->GetRock();
			return pMapChannel;
		}
		pPrevChannel = pMapChannel;
	}
	return NULL;
}

/////////////////////////////////////////////////////////////////////////////
// Class for loaded sounds
/////////////////////////////////////////////////////////////////////////////

CWinGlkSound::CWinGlkSound(LPCTSTR pszFileName)
{
	m_bDelete = true;
	m_pData = NULL;
	m_iLength = 0;

	m_iRepeat = 0;

	CFile SoundFile;
	if (SoundFile.Open(pszFileName,CFile::modeRead))
	{
		m_iLength = SoundFile.GetLength();
		m_pData = new BYTE[m_iLength];
		SoundFile.Read(m_pData,m_iLength);
	}
}

CWinGlkSound::CWinGlkSound(BYTE* pData, int iLength)
{
	m_bDelete = false;
	m_pData = pData;
	m_iLength = iLength;

	m_iRepeat = 0;
}

CWinGlkSound::~CWinGlkSound()
{
	if (m_bDelete)
		delete[] m_pData;
}

double CWinGlkSound::DecibelVolume(int iVolume)
{
	// The volume argument is from 0 to 0x10000, with 0x8000
	// representing half volume. This is converted to a decibel
	// level, with -10dB representing half volume, -20dB one
	// quarter volume, and so on.
	// Zero volume is taken to be -100dB.

	if (iVolume < 0)
		iVolume = 0;
	if (iVolume > 0x10000)
		iVolume = 0x10000;

	double dDecibels;
	if (iVolume == 0)
		dDecibels = -100.0;
	else
		dDecibels = -10.0 * (log((double)0x10000 / (double)iVolume) / log(2));

	if (dDecibels > 0.0)
		dDecibels = 0.0;
	if (dDecibels < -100.0)
		dDecibels = -100.0;
	return dDecibels;
}

/////////////////////////////////////////////////////////////////////////////
// Base class for sound loaders
/////////////////////////////////////////////////////////////////////////////

// Set up the loaders
void CWinGlkSoundLoader::InitLoaders(void)
{
	m_Loaders.Add(new CWinGlkAIFFSoundLoader());
	m_Loaders.Add(new CWinGlkMODSoundLoader());
}

	// Notify loaders to clean up after sound playback has been stopped
void CWinGlkSoundLoader::AllSoundStopped(void)
{
	for (int i = 0; i < m_Loaders.GetSize(); i++)
		m_Loaders[i]->SoundStopped();
}

// Delete the loaders
void CWinGlkSoundLoader::RemoveLoaders(void)
{
	for (int i = 0; i < m_Loaders.GetSize(); i++)
		delete m_Loaders[i];
	m_Loaders.RemoveAll();
}

// Get a loader for a given sound identifier
CWinGlkSoundLoader* CWinGlkSoundLoader::GetLoaderForID(glui32 id)
{
	for (int i = 0; i < GetLoaderCount(); i++)
	{
		CWinGlkSoundLoader* pLoader = GetLoader(i);
		if (pLoader->GetIdentifier() == id)
			return pLoader;
	}
	return NULL;
}

CArray<CWinGlkSoundLoader*,CWinGlkSoundLoader*> CWinGlkSoundLoader::m_Loaders;
