/////////////////////////////////////////////////////////////////////////////
//
// Windows MFC Glk Libraries
//
// GlkSoundAIFF
// Glk AIFF sounds
//
/////////////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "GlkDll.h"
#include "GlkSoundAIFF.h"

#include <math.h>

extern "C"
{
#include "gi_blorb.h"
}

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// Class for AIFF sound loader
/////////////////////////////////////////////////////////////////////////////

// Get file prefix for sounds supported by this loader
LPCTSTR CWinGlkAIFFSoundLoader::GetFilePrefix(void)
{
	return "snd";
}

// Get the file extensions for sounds supported for this loader
int CWinGlkAIFFSoundLoader::GetNumberFileExtensions(void)
{
	return 2;
}

LPCTSTR CWinGlkAIFFSoundLoader::GetFileExtension(int iExtIndex)
{
	switch (iExtIndex)
	{
	case 0:
		return "aif";
	case 1:
		return "aiff";
	}
	return "";
}

// Get the identifier for sounds supported for this loader
glui32 CWinGlkAIFFSoundLoader::GetIdentifier(void)
{
	return giblorb_make_id('F','O','R','M');
}

// Get a sound object
CWinGlkSound* CWinGlkAIFFSoundLoader::GetSound(LPCTSTR pszFileName)
{
	return new CWinGlkAIFFSound(pszFileName);
}

CWinGlkSound* CWinGlkAIFFSoundLoader::GetSound(BYTE* pData, int iLength)
{
	return new CWinGlkAIFFSound(pData,iLength);
}

/////////////////////////////////////////////////////////////////////////////
// Class for AIFF sounds
/////////////////////////////////////////////////////////////////////////////

CWinGlkAIFFSound::CWinGlkAIFFSound(BYTE* pData, int iLength) : CWinGlkSound(pData,iLength)
{
	m_pRenderPtr = NULL;
	m_pRenderMin = NULL;
	m_pRenderMax = NULL;
	m_Duration = 0;
}

CWinGlkAIFFSound::CWinGlkAIFFSound(LPCTSTR pszFileName) : CWinGlkSound(pszFileName)
{
	m_pRenderPtr = NULL;
	m_pRenderMin = NULL;
	m_pRenderMax = NULL;
	m_Duration = 0;
}

CWinGlkAIFFSound::~CWinGlkAIFFSound()
{
	RemoveFromList();
}

bool CWinGlkAIFFSound::Play(int iRepeat, int iVolume)
{
	if (m_pData == NULL)
		return false;

	// Check for AIFF header
	if (strncmp((char*)m_pData,"FORM",4) != 0)
		return false;
	if (strncmp((char*)(m_pData+8),"AIFF",4) != 0)
		return false;

	// Find the COMM chunk
	BYTE* pData = FindChunk("COMM");
	if (pData == NULL)
		return false;

	// Read in details of the sample
	unsigned short Channels = ReadShort(pData);
	unsigned long Samples = ReadLong(pData+2);
	unsigned short Bits = ReadShort(pData+6);
	double Rate = ReadExtended(pData+8);

	// Find the SSND chunk
	pData = FindChunk("SSND");
	if (pData == NULL)
		return false;

	// Create a buffer
	if (CreateBuffer(Channels,(int)Rate,Bits) == false)
		return false;

	// Set the duration of the sample
	if (iRepeat > 0)
		m_Duration = (DWORD)ceil((Samples * iRepeat * 1000.0) / Rate);
	else
		m_Duration = -1;

	// Set up the current position for rendering wave data
	m_pRenderPtr = pData;
	m_pRenderMin = m_pRenderPtr;
	m_pRenderMax = pData + ((Bits>>3)*Samples*Channels);

	// Fill the buffer with sample data
	m_iRepeat = (iRepeat < 0) ? -1 : iRepeat - 1;
	if (FillBuffer(GetBufferSize()) == false)
		return false;

	// Set the volume for the buffer
	SetVolume(iVolume);

	// Start the buffer playing
	return PlayBuffer();
}

bool CWinGlkAIFFSound::IsPlaying(void)
{
	return IsBufferPlaying();
}

void CWinGlkAIFFSound::SetVolume(int iVolume)
{
	// The SetVolume() call to DirectSound requires a volume
	// in 100ths of a decibel.
	SetBufferVolume((LONG)DecibelVolume(iVolume) * 100L);
}

// Write sample data into the supplied PCM sample buffers
void CWinGlkAIFFSound::WriteSampleData(unsigned char* pSample, int iSampleLen)
{
	switch (m_Format.wBitsPerSample)
	{
	case 8:
		{
			for (int i = 0; i < iSampleLen; i++)
			{
				if (CheckRenderPtr())
					*(pSample++) = (unsigned char)((*(m_pRenderPtr++)) ^ 0x80);
				else
					*(pSample++) = 0x80;
			}
		}
		break;
	case 16:
		{
			unsigned short* pWritePtr = (unsigned short*)pSample;
			for (int i = 0; i < iSampleLen; i += 2)
			{
				if (CheckRenderPtr())
				{
					*(pWritePtr++) = (unsigned short)
						(((*m_pRenderPtr)<<8) | *(m_pRenderPtr+1));
					m_pRenderPtr += 2;
				}
				else
					*(pWritePtr++) = 0;
			}
		}
		break;
	}
}

// Test that the current point into the AIFF buffer is valid
bool CWinGlkAIFFSound::CheckRenderPtr(void)
{
	if (m_pRenderPtr >= m_pRenderMax)
	{
		// Fail if this is the end of the last repeat
		if (m_iRepeat == 0)
			return false;

		// If not looping forever, decrement the repeat counter
		if (m_iRepeat > 0)
			m_iRepeat--;

		// Reset the pointer
		m_pRenderPtr = m_pRenderMin;
	}
	return true;
}

// Check if the sound has finished playing
bool CWinGlkAIFFSound::IsSoundOver(DWORD Tick)
{
	if (m_Playing == false)
		return true;

	// Check if sound is playing forever
	if (m_Duration < 0)
		return false;
	return (Tick > m_StartTime + m_Duration);
}

// Get a type identifier for the sound
int CWinGlkAIFFSound::GetType(void)
{
	return (int)'A';
}

// Find an AIFF chunk
BYTE* CWinGlkAIFFSound::FindChunk(LPCTSTR pszChunk)
{
	BYTE* pData = m_pData+12;
	while (true)
	{
		if (strncmp((char*)pData,pszChunk,4) == 0)
			return pData+8;

		// Move to the next chunk
		pData += ReadLong(pData+4)+8;

		if ((pData - m_pData) >= m_iLength)
			break;
	}
	return NULL;
}

unsigned short CWinGlkAIFFSound::ReadShort(const unsigned char *bytes)
{
	return (unsigned short)(
		((unsigned short)(bytes[0] & 0xFF) << 8) |
		((unsigned short)(bytes[1] & 0xFF)));
}

unsigned long CWinGlkAIFFSound::ReadLong(const unsigned char *bytes)
{
	return (unsigned long)(
		((unsigned long)(bytes[0] & 0xFF) << 24) |
		((unsigned long)(bytes[1] & 0xFF) << 16) |
		((unsigned long)(bytes[2] & 0xFF) << 8) |
		((unsigned long)(bytes[3] & 0xFF)));
}

/* 
 * Copyright (C) 1988-1991 Apple Computer, Inc.
 * All rights reserved.
 *
 * Machine-independent I/O routines for IEEE floating-point numbers.
 *
 * NaN's and infinities are converted to HUGE_VAL or HUGE, which
 * happens to be infinity on IEEE machines.  Unfortunately, it is
 * impossible to preserve NaN's in a machine-independent way.
 * Infinities are, however, preserved on IEEE machines.
 *
 * These routines have been tested on the following machines:
 *    Apple Macintosh, MPW 3.1 C compiler
 *    Apple Macintosh, THINK C compiler
 *    Silicon Graphics IRIS, MIPS compiler
 *    Cray X/MP and Y/MP
 *    Digital Equipment VAX
 *
 * Implemented by Malcolm Slaney and Ken Turkowski.
 *
 * Malcolm Slaney contributions during 1988-1990 include big- and little-
 * endian file I/O, conversion to and from Motorola's extended 80-bit
 * floating-point format, and conversions to and from IEEE single-
 * precision floating-point format.
 *
 * In 1991, Ken Turkowski implemented the conversions to and from
 * IEEE double-precision format, added more precision to the extended
 * conversions, and accommodated conversions involving +/- infinity,
 * NaN's, and denormalized numbers.
 */
#define UnsignedToFloat(u) (((double)((long)(u - 2147483647L - 1))) + 2147483648.0)
double CWinGlkAIFFSound::ReadExtended(const unsigned char *bytes)
{
	double f;
	int expon;
	unsigned long hiMant, loMant;

	expon = ((bytes[0] & 0x7F) << 8) | (bytes[1] & 0xFF);
	hiMant = ReadLong(bytes+2);
	loMant = ReadLong(bytes+6);

	if (expon == 0 && hiMant == 0 && loMant == 0)
		f = 0;
	else
	{
		if (expon == 0x7FFF) /* Infinity or NaN */
			f = -1;
		else
		{
			expon -= 16383;
			f = ldexp(UnsignedToFloat(hiMant),expon -= 31);
			f += ldexp(UnsignedToFloat(loMant),expon -= 32);
		}
	}

	if (bytes[0] & 0x80)
		return -f;
	return f;
}
